/**
 * Copyright (C) 2011 Angelo Zerr <angelo.zerr@gmail.com> and Pascal Leclercq <pascal.leclercq@gmail.com>
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free  of charge, to any person obtaining
 * a  copy  of this  software  and  associated  documentation files  (the
 * "Software"), to  deal in  the Software without  restriction, including
 * without limitation  the rights to  use, copy, modify,  merge, publish,
 * distribute,  sublicense, and/or sell  copies of  the Software,  and to
 * permit persons to whom the Software  is furnished to do so, subject to
 * the following conditions:
 *
 * The  above  copyright  notice  and  this permission  notice  shall  be
 * included in all copies or substantial portions of the Software.
 *
 * THE  SOFTWARE IS  PROVIDED  "AS  IS", WITHOUT  WARRANTY  OF ANY  KIND,
 * EXPRESS OR  IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE  WARRANTIES OF
 * MERCHANTABILITY,    FITNESS    FOR    A   PARTICULAR    PURPOSE    AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE,  ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package fr.opensagres.xdocreport.core.registry;

import static fr.opensagres.xdocreport.core.utils.StringUtils.isNotEmpty;

import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;

import fr.opensagres.xdocreport.core.debugger.IRegistryDebugger;

public abstract class AbstractRegistry {

	private static final String CLASS_PROPERTY = "class";

	private boolean initialized;
	private final String propertiesFileName;
	private final String registryType;
	private IRegistryDebugger debugger;

	public AbstractRegistry(String propertiesFileName, String registryType) {
		this.propertiesFileName = propertiesFileName;
		this.registryType = registryType;
	}

	public void initialize() {
		initializeIfNeeded();
	}
	
	protected void initializeIfNeeded() {
		initializeIfNeeded(debugger);
	}

	protected void initializeIfNeeded(IRegistryDebugger debugger) {
		if (initialized) {
			// registry already initialized.
			return;
		}
		onStartInitialization();
		// Start registry initialization.
		long startTime = -1;
		if (isDebuggable(debugger)) {
			startTime = System.currentTimeMillis();
			debugger.startRegistryInitialization(registryType,
					propertiesFileName);
		}
		try {
			// Get list of properties file found in the classloader.
			Enumeration<URL> urls = getResources(getClass().getClassLoader(),
					propertiesFileName);
			if (urls != null) {
				// Loop for properties file found.
				while (urls.hasMoreElements()) {
					URL url = urls.nextElement();
					Properties p = new Properties();
					try {
						// Start loading current property file.
						if (isDebuggable(debugger)) {
							debugger.startLoadingPropertiesFile(url);
						}
						p.load(url.openStream());

						// Loop for each property of the current file.
						Set<Entry<Object, Object>> propertiesSet = p.entrySet();
						for (Entry<Object, Object> entry : propertiesSet) {
							String key = (String) entry.getKey();
							if (key != null && key.startsWith(CLASS_PROPERTY)) {
								// property which start 'class' found,
								// instanciate the class and process it.
								String clazz = (String) entry.getValue();
								if (isNotEmpty(clazz)) {
									try {
//										Object instance = Class.forName(clazz)
//												.newInstance();
										//PLQ  : this is more OSGi compliant
										Object instance = getClass().getClassLoader().loadClass(clazz).newInstance();
										boolean result = registerInstance(instance);
										if (result) {
											if (isDebuggable(debugger)) {
												debugger.addReportFactoryForReportLoader(clazz);
											}
										} else {
											if (isDebuggable(debugger)) {
												debugger.dontImplementClass(clazz);
											}
										}
									} catch (Throwable e) {
										// Error while instantiate class.
										if (isDebuggable(debugger)) {
											debugger.cannotInstantiateClass(
													clazz, e);
										}
									}
								}
							}
						}
						if (isDebuggable(debugger)) {
							// End loading current property file with no error
							debugger.endLoadingPropertiesFile(url, null);
						}
					} catch (IOException e) {
						// End loading current property file with error
						if (isDebuggable(debugger)) {
							debugger.endLoadingPropertiesFile(url, e);
						}
					}
				}
			}
		} catch (IOException e1) {
			// End registry initialization with error.
			if (isDebuggable(debugger)) {
				debugger.endRegistryInitialization(registryType,
						System.currentTimeMillis() - startTime, e1);
			}

		} finally {
			// End registry initialization with no error.
			initialized = true;
			if (isDebuggable(debugger)) {
				debugger.endRegistryInitialization(registryType,
						System.currentTimeMillis() - startTime, null);
			}
			onEndInitialization();
		}
	}

	/**
	 * Method called when registry start initialization.
	 */
	protected void onStartInitialization() {

	}

	/**
	 * Method called when registry end initialization.
	 */
	protected void onEndInitialization() {

	}

	/**
	 * Returns a resources founded from the ClassLoader.
	 * 
	 * @param cl
	 * @param resourcePath
	 * @return
	 * @throws IOException
	 */
	private static Enumeration<URL> getResources(ClassLoader cl,
			String resourcePath) throws IOException {
		return cl == null ? ClassLoader.getSystemResources(resourcePath) : cl
				.getResources(resourcePath);
	}

	/**
	 * Returns true if registry initialization can be tracked and false
	 * otherwise.
	 * 
	 * @param debugger
	 * @return
	 */
	private boolean isDebuggable(IRegistryDebugger debugger) {
		return debugger != null;
	}

	/**
	 * returns registry debugger.
	 * 
	 * @return
	 */
	public IRegistryDebugger getDebugger() {
		return debugger;
	}

	/**
	 * Set registry debugger.
	 * 
	 * @param debuuger
	 */
	public void setDebugger(IRegistryDebugger debugger) {
		this.debugger = debugger;
	}

	/**
	 * Register the instance in the registry.
	 * 
	 * @param instance
	 * @return true if instance cannot be registered and false otherwise.
	 */
	protected abstract boolean registerInstance(Object instance);

	/**
	 * Returns true if registry is initialized and false otherwise.
	 * 
	 * @return
	 */
	public boolean isInitialized() {
		return initialized;
	}

	/**
	 * Dispose the registry.
	 */
	public void dispose() {
		doDispose();
		this.initialized = false;
	}

	protected abstract void doDispose();
}
